package info.textgrid.middleware;

import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.net.URLDecoder;
import java.text.ParseException;
import java.util.Hashtable;
import java.util.Map;
import javax.xml.datatype.DatatypeConfigurationException;
import org.apache.commons.logging.LogFactory;
import org.elasticsearch.action.get.GetRequest;
import org.elasticsearch.action.get.GetResponse;
import org.elasticsearch.action.search.SearchRequest;
import org.elasticsearch.action.search.SearchResponse;
import org.elasticsearch.action.search.SearchScrollRequest;
import org.elasticsearch.client.RequestOptions;
import org.elasticsearch.client.RestHighLevelClient;
import org.elasticsearch.common.Strings;
import org.elasticsearch.common.unit.TimeValue;
import org.elasticsearch.index.query.QueryBuilder;
import org.elasticsearch.index.query.QueryBuilders;
import org.elasticsearch.search.SearchHit;
import org.elasticsearch.search.builder.SearchSourceBuilder;
import org.elasticsearch.search.fetch.subphase.FetchSourceContext;
import org.json.JSONObject;
import info.textgrid.middleware.oaipmh.HeaderType;
import info.textgrid.middleware.oaipmh.ListRecordsType;
import info.textgrid.middleware.oaipmh.MetadataType;
import info.textgrid.middleware.oaipmh.RecordType;
import info.textgrid.middleware.oaipmh.ResumptionTokenType;

/**
 *
 */
public class RecordListDelivererDC extends RecordListDelivererAbstract {

  // **
  // FINALS
  // **

  private static final int LIFETIME_RES_TOKEN = 600;

  // **
  // STATICS
  // **

  private static org.apache.commons.logging.Log log = LogFactory.getLog(OAIPMHImpl.class);
  protected static Map<String, Integer> cursorCollector = new Hashtable<String, Integer>();

  // **
  // DC-Field Lists
  // **

  private String[] contributorList;
  private String[] coverageList;
  private String[] creatorList;
  private String[] dateList;
  private String[] descriptionList;
  private String[] formatList;
  private String[] identifierList;
  private String[] languageList;
  private String[] publisherList;
  private String[] relationList;
  private String[] relationsForWorkList;
  private String[] rightsList;
  private String[] sourceList;
  private String[] subjectList;
  private String[] titleList;
  private String[] typeList;

  /**
   * @param textgrid
   * @param dariah
   */
  public RecordListDelivererDC(boolean textgrid, boolean dariah) {
    super(textgrid, dariah);
  }

  /**
   * <p>
   * The furtherDCElementes fetches the specified fields from the work object. This function is
   * called from the function fetchFields.
   * </p>
   * 
   * @param id
   * @param client
   * @return
   */
  public GetResponse furtherDCElements(String id, RestHighLevelClient client) {

    if (id != null) {
      try {
        id = URLDecoder.decode(id, "UTF-8");
      } catch (UnsupportedEncodingException e) {
        e.printStackTrace();
      }
      String[] includes = this.workFields;
      String[] excludes = Strings.EMPTY_ARRAY;
      FetchSourceContext fetchSourceContext = new FetchSourceContext(true, includes, excludes);
      GetRequest getRequest =
          new GetRequest(OAI_ESClient.getEsIndex(), OAI_ESClient.getEsType(), id)
              .fetchSourceContext(fetchSourceContext);

      GetResponse responseWorkValues = null;
      try {
        responseWorkValues = OAI_ESClient.getEsClient().get(getRequest, RequestOptions.DEFAULT);
      } catch (IOException e) {
        // TODO Auto-generated catch block
        e.printStackTrace();
      }

      return responseWorkValues;
    }

    return null;
  }

  /**
   * <p>
   * The fetchFields function filters all edition and fetch the specified fields
   * </p>
   * 
   * @param query
   * @param recordList
   * @param resumptionToken
   * @param set
   * @throws UnsupportedEncodingException
   */
  public void fetchFields(QueryBuilder query, ListRecordsType recordList, String resumptionToken,
      String set) throws UnsupportedEncodingException {

    String scrollID = "";
    SearchResponse scrollResp;
    QueryBuilder recordFilter;
    DublinCoreBuilder dublinCoreBuilder = new DublinCoreBuilder();

    log.debug("formatField: " + (this.formatField == null ? "NULL" : this.formatField));
    log.debug("formatToFilter: " + (this.formatToFilter == null ? "NULL" : this.formatToFilter));

    if (this.dariah == true) {
      // All objects!
      recordFilter = query;
    } else {
      // All editions only!
      recordFilter = QueryBuilders.boolQuery().must(query)
          .must(QueryBuilders.matchPhraseQuery(this.formatField, this.formatToFilter));
    }

    SearchRequest searchRequest = new SearchRequest(OAI_ESClient.getEsIndex());
    SearchSourceBuilder searchSourceBuilder = new SearchSourceBuilder();

    scrollResp = null;

    searchSourceBuilder.query(recordFilter);
    searchSourceBuilder.size(100);
    searchRequest.source(searchSourceBuilder);

    if (resumptionToken != null) {
      SearchScrollRequest scrollRequest = new SearchScrollRequest(resumptionToken);
      scrollRequest.scroll(TimeValue.timeValueSeconds(LIFETIME_RES_TOKEN));

      try {
        scrollResp = OAI_ESClient.getEsClient().scroll(scrollRequest, RequestOptions.DEFAULT);
      } catch (IOException e) {
        // TODO Auto-generated catch block
        e.printStackTrace();
      }
    } else {
      searchRequest.source(searchSourceBuilder);
      searchRequest.scroll(TimeValue.timeValueMinutes(LIFETIME_RES_TOKEN));
      try {
        scrollResp = OAI_ESClient.getEsClient().search(searchRequest, RequestOptions.DEFAULT);
      } catch (IOException e) {
        // TODO Auto-generated catch block
        e.printStackTrace();
      }
    }

    log.info("Queried fields: " + this.fields.toString());
    log.info("Lifetime for resumption token is set to: " + LIFETIME_RES_TOKEN);

    scrollID = scrollResp.getScrollId();

    long completeListSize = scrollResp.getHits().totalHits;
    setResultSize(completeListSize);

    if (completeListSize > 0) {
      setFoundItems(true);

      int i = 0;
      for (SearchHit hit : scrollResp.getHits().getHits()) {
        i++;

        if (hit != null && hit.getFields() != null) {
          try {
            if (hit.getSourceAsMap().get(this.dateOfObjectCreation) != null) {
              String datestamp = hit.getSourceAsMap().get(this.dateOfObjectCreation).toString();
              datestamp = OAIPMHUtilities.convertDateFormat(datestamp).toXMLFormat();
            }
          } catch (ParseException ee) {
            log.debug(ee);
          } catch (DatatypeConfigurationException e) {
            log.debug(e);
          }

          // ES6
          JSONObject json = new JSONObject(hit.getSourceAsMap());
          this.modifiedValue = OAIPMHUtilities.fieldLoader(json, this.modifiedField);
          // OLD
          // this.modifiedValue = hit.getSourceAsMap().get(this.modifiedField).toString();

          // TextGrid search!
          if (this.textgrid) {
            String workUri = "";
            String[] blaFields = new String[] {this.relationToFurtherMetadataObject};

            if (hit.getSourceAsMap().get(this.formatField).toString().equals(this.formatToFilter)) {
              String setSpec = "project:"
                  + DublinCoreFieldLoader.fillList(hit, TGConstants.RELATIONS_LIST).get(0);
              if (DublinCoreFieldLoader.fillList(hit, blaFields).get(0) != null) {
                workUri = DublinCoreFieldLoader.fillList(hit, blaFields).get(0)
                    .substring(this.repositoryObjectURIPrefix.length());

                if (!workUri.endsWith(".0")) {
                  workUri = workUri.concat(".0");
                }
              }
              try {
                dublinCoreBuilder = putContentIntoDCFieldLists(hit,
                    furtherDCElements(workUri, OAI_ESClient.getEsClient()));
              } catch (ParseException e) {
                // TODO Auto-generated catch block
                e.printStackTrace();
              } catch (DatatypeConfigurationException e) {
                // TODO Auto-generated catch block
                e.printStackTrace();
              }

              log.debug(
                  "ID field: " + (this.identifierField == null ? "NULL" : this.identifierField));

              String identifier = hit.getSourceAsMap().get(this.identifierField).toString();

              log.debug("identifier (tgrep): " + identifier);

              buildRecord(recordList, set, identifier, dublinCoreBuilder, setSpec);
            }
          }

          // DARIAH search!
          else if (this.dariah) {

            // Get identifier field.
            String identifier = OAIPMHUtilities.fieldLoader(json, this.identifierField);
            String setSpec = identifier;

            log.debug("identifier (dhrep): " + identifier);
            log.debug("setSpec: " + setSpec);

            try {
              dublinCoreBuilder = putContentIntoDCFieldLists2(hit,
                  furtherDCElements(hit.getId(), OAI_ESClient.getEsClient()));
            } catch (ParseException e) {
              // TODO Auto-generated catch block
              e.printStackTrace();
            } catch (DatatypeConfigurationException e) {
              // TODO Auto-generated catch block
              e.printStackTrace();
            }

            buildRecord(recordList, set, identifier, dublinCoreBuilder, setSpec);
          }
        }
      }

      // Check the need for a resumption token!
      ResumptionTokenType resTokenForResponse = OAIPMHUtilities.getResumptionToken(completeListSize,
          resumptionToken, cursorCollector, scrollID, this.searchResponseSize, i);

      // if (resTokenForResponse != null) {

      recordList.setResumptionToken(resTokenForResponse);

      log.debug("RESTOKEN: ");
      log.debug(recordList.getResumptionToken());

      // }

    } else {
      setFoundItems(false);
    }
  }

  /*
   * public DublinCoreBuilder putContentIntoDCFieldLists(JSONObject json) {
   * 
   * // DublinCoreBuilder result = new DublinCoreBuilder();
   * //result.setContributor(DublinCoreFieldLoader.fillList(responseWorkValues,
   * this.contributorList)); /*result.setCoverage(DublinCoreFieldLoader.fillList(responseWorkValues,
   * this.coverageList)); result.setCreator(DublinCoreFieldLoader.fillList(responseWorkValues,
   * this.creatorList)); result.setDate(DublinCoreFieldLoader.fillList(responseWorkValues,
   * this.dateList));
   * 
   * result.setDescription( DublinCoreFieldLoader.fillList(responseWorkValues,
   * this.descriptionList));
   */
  // result.setFormat(DublinCoreFieldLoader.fillList(hit, this.formatList));

  // result.setIdentifier(OAIPMHUtilities.fieldLoader(json, fields));
  // result.setLanguage(DublinCoreFieldLoader.fillList(hit, this.languageList));
  // result.setPublisher(DublinCoreFieldLoader.fillList(hit, this.publisherList));
  // result.setRelation(DublinCoreFieldLoader.fillList(hit, this.relationList));
  // result.setRelation(
  // DublinCoreFieldLoader.fillList(responseWorkValues, this.relationsForWorkList));
  // result.setRights(DublinCoreFieldLoader.fillList(hit, this.rightList));
  // result.setSource(DublinCoreFieldLoader.fillList(hit, this.sourceList));
  // result.setSubject(DublinCoreFieldLoader.fillList(hit, this.subjectList));
  // result.setTitle(DublinCoreFieldLoader.fillList(hit, this.titleList));
  // result.setType(DublinCoreFieldLoader.fillList(responseWorkValues, this.typeList));

  // return result;
  // }*/

  /**
   * @param hit
   * @param responseWorkValues
   * @return
   * @throws DatatypeConfigurationException
   * @throws ParseException
   */
  public DublinCoreBuilder putContentIntoDCFieldLists(SearchHit hit,
      GetResponse responseWorkValues) throws ParseException, DatatypeConfigurationException {

    DublinCoreBuilder result = new DublinCoreBuilder();

    result.setContributor(DublinCoreFieldLoader.fillList(responseWorkValues, this.contributorList));
    result.setCoverage(DublinCoreFieldLoader.fillList(responseWorkValues, this.coverageList));
    result.setCreator(DublinCoreFieldLoader.fillList(responseWorkValues, this.creatorList));
    result.setDate(DublinCoreFieldLoader.fillList(responseWorkValues, this.dateList));

    result.setDescription(
        DublinCoreFieldLoader.fillList(responseWorkValues, this.descriptionList));
    result.setFormat(DublinCoreFieldLoader.fillList(hit, this.formatList));
    // result.setIdentifier(DublinCoreFieldLoader.fillList(hit, this.identifierList));

    log.debug("hit size: " + hit.getFields().size());
    log.debug("hit id: " + hit.getId());
    log.debug("hit map keys: " + hit.getSourceAsMap().keySet());
    log.debug("hit map vals: " + hit.getSourceAsMap().entrySet());
    log.debug("identifier list: " + this.identifierList);

    result.setIdentifier(
        OAIPMHUtilities.fieldLoader(new JSONObject(hit.getSourceAsMap()), this.identifierList));
    result.setLanguage(DublinCoreFieldLoader.fillList(hit, this.languageList));
    result.setPublisher(DublinCoreFieldLoader.fillList(hit, this.publisherList));
    result.setRelation(DublinCoreFieldLoader.fillList(hit, this.relationList));
    result.setRelation(
        DublinCoreFieldLoader.fillList(responseWorkValues, this.relationsForWorkList));
    result.setRights(DublinCoreFieldLoader.fillList(hit, this.rightsList));
    result.setSource(DublinCoreFieldLoader.fillList(hit, this.sourceList));
    result.setSubject(DublinCoreFieldLoader.fillList(hit, this.subjectList));
    result.setTitle(DublinCoreFieldLoader.fillList(hit, this.titleList));
    result.setType(DublinCoreFieldLoader.fillList(responseWorkValues, this.typeList));

    return result;
  }

  /**
   * @param hit
   * @param responseWorkValues
   * @return
   * @throws DatatypeConfigurationException
   * @throws ParseException
   */
  public DublinCoreBuilder putContentIntoDCFieldLists2(SearchHit hit,
      GetResponse responseWorkValues) throws ParseException, DatatypeConfigurationException {

    DublinCoreBuilder result = new DublinCoreBuilder();

    result.setContributor(DublinCoreFieldLoader.fillList(hit, this.contributorList));
    result.setCoverage(DublinCoreFieldLoader.fillList(hit, this.coverageList));
    result.setCreator(DublinCoreFieldLoader.fillList(hit, this.creatorList));
    result.setDate(DublinCoreFieldLoader.fillList(hit, this.dateList));
    result.setDescription(DublinCoreFieldLoader.fillList(hit, this.descriptionList));
    result.setFormat(DublinCoreFieldLoader.fillList(hit, this.formatList));
    // result.setIdentifier(DublinCoreFieldLoader.fillList(hit, this.identifierList));
    result.setIdentifier(
        OAIPMHUtilities.fieldLoader(new JSONObject(hit.getSourceAsMap()), this.identifierList));
    result.setLanguage(DublinCoreFieldLoader.fillList(hit, this.languageList));
    result.setPublisher(DublinCoreFieldLoader.fillList(hit, this.publisherList));
    result.setRelation(DublinCoreFieldLoader.fillList(hit, this.relationList));
    result.setRights(DublinCoreFieldLoader.fillList(hit, this.rightsList));
    result.setSource(DublinCoreFieldLoader.fillList(hit, this.sourceList));
    result.setSubject(DublinCoreFieldLoader.fillList(hit, this.subjectList));
    result.setTitle(DublinCoreFieldLoader.fillList(hit, this.titleList));
    result.setType(DublinCoreFieldLoader.fillList(hit, this.typeList));

    return result;
  }

  /**
   * @param recordList
   * @param set
   * @param headerIdentifier
   * @param dublinCoreBuilder
   * @return
   */
  public RecordType buildRecord(ListRecordsType recordList, String set, String headerIdentifier,
      DublinCoreBuilder dublinCoreBuilder, String setSpec) {

    MetadataType metadata = new MetadataType();
    RecordType record = new RecordType();

    log.debug("headerIdentifier: " + headerIdentifier);

    metadata = dublinCoreBuilder.getDC();
    record.setMetadata(metadata);
    record.setHeader(setHeader(set, headerIdentifier, setSpec));
    recordList.getRecord().add(record);

    return record;
  }

  /*
   * (non-Javadoc)
   * 
   * @see info.textgrid.middleware.RecordListDelivererInterface#getRecords(java.lang.String,
   * java.lang.String, java.lang.String, java.lang.String)
   */
  public ListRecordsType getRecords(String from, String to, String set, String resumptionToken) {

    log.info("Started List Records with booleans for DARIAH:  " + this.dariah + " and TextGrid: "
        + this.textgrid);

    ListRecordsType recordList = new ListRecordsType();
    QueryBuilder query;
    QueryBuilder rangeQuery = QueryBuilders.rangeQuery(this.rangeField).from(from).to(to);

    if (set != null) {
      String queryField = "";
      String valueField = "";

      if (this.dariah == true) {
        queryField = "descriptiveMetadata.dc:relation";
        valueField = set;
      }

      if (this.textgrid == true) {
        String[] setParts = set.split(":");
        if (setParts[0].equals("project")) {
          queryField = "project.id";
          valueField = setParts[1];
        }
      }

      QueryBuilder matchQuery = QueryBuilders.matchPhraseQuery(queryField, valueField);
      QueryBuilder boolQuery = QueryBuilders.boolQuery().must(rangeQuery).must(matchQuery);

      log.debug("boolQuery: " + boolQuery);

      query = boolQuery;
    } else {
      query = rangeQuery;
    }

    try {

      log.debug("getRecords(): " + (query != null ? "queryName=" + query.getName() : "query=NULL")
          + " | " + (recordList != null ? recordList : "recordList=NULL") + " | "
          + (resumptionToken != null ? resumptionToken : "resumptionToken=NULL") + " | "
          + (set != null ? set : "set=NULL"));

      fetchFields(query, recordList, resumptionToken, set);

    } catch (UnsupportedEncodingException e) {
      e.printStackTrace();
    }

    return recordList;
  }

  /**
   * @param set
   * @param headerIdentifier
   * @return
   */
  public HeaderType setHeader(final String set, final String headerIdentifier, String setSpec) {

    HeaderType header = new HeaderType();
    String identifierForHeader = headerIdentifier;

    // Set date in XML format.
    try {
      header.setDatestamp(OAIPMHUtilities.convertDateFormat(this.modifiedValue).toXMLFormat());
    } catch (ParseException e) {
      // TODO Auto-generated catch block
      e.printStackTrace();
    } catch (DatatypeConfigurationException e) {
      // TODO Auto-generated catch block
      e.printStackTrace();
    }

    // Set header specific for TextGrid objects (remove URL).
    if (this.textgrid == true) {
      identifierForHeader = identifierForHeader.replaceFirst("https://textgridrep.org/", "");
    }

    header.setIdentifier(identifierForHeader);

    // Set set :-)
    if (set != null) {
      header.getSetSpec().add(setSpec);
    }

    return header;
  }

  /*
   * Setter for DublinCore Lists
   */

  /**
   * @param contributorList
   */
  public void setContributors(String[] contributorList) {
    this.contributorList = contributorList;
  }

  /**
   * @param coveragesList
   */
  public void setCoverages(String[] coveragesList) {
    this.coverageList = coveragesList;
  }

  /**
   * @param creatorsList
   */
  public void setCreators(String[] creatorsList) {
    this.creatorList = creatorsList;
  }

  /**
   * @param datesList
   */
  public void setDates(String[] datesList) {
    this.dateList = datesList;
  }

  /**
   * @param descriptionsList
   */
  public void setDescriptions(String[] descriptionsList) {
    this.descriptionList = descriptionsList;
  }

  /**
   * @param formatsList
   */
  public void setFormats(String[] formatsList) {
    this.formatList = formatsList;
  }

  /**
   * @param identifiersList
   */
  public void setIdentifiers(String[] identifiersList) {
    this.identifierList = identifiersList;
  }

  /**
   * @param languagesList
   */
  public void setLanguages(String[] languagesList) {
    this.languageList = languagesList;
  }

  /**
   * @param publishersList
   */
  public void setPublishers(String[] publishersList) {
    this.publisherList = publishersList;
  }

  /**
   * @param rightsList
   */
  public void setRights(String[] rightsList) {
    this.rightsList = rightsList;
  }

  /**
   * @param sourcesList
   */
  public void setSources(String[] sourcesList) {
    this.sourceList = sourcesList;
  }

  /**
   * @param subjectsList
   */
  public void setSubjects(String[] subjectsList) {
    this.subjectList = subjectsList;
  }

  /**
   * @param titlesList
   */
  public void setTitles(String[] titlesList) {
    this.titleList = titlesList;
  }

  /**
   * @param typesList
   */
  public void setTypes(String[] typesList) {
    this.typeList = typesList;
  }

  /**
   * @param relationsList
   */
  public void setRelations(String[] relationsList) {
    this.relationList = relationsList;
  }

  /**
   * @param relationsForWorkList
   */
  public void setRelationsForWork(String[] relationsForWorkList) {
    this.relationsForWorkList = relationsForWorkList;
  }

}
